#!/usr/bin/python -u
'''
Tahmo Bridge for BayEOS 
- Reads observation data in json format from tahmo server using http interface  
- Saves data in BayEOS labeled frame format by bayeosgatewayclient 
@author: oliver.archner@uni-bayreuth.de
Created on 05.07.2017
'''

from time import sleep
from bayeosgatewayclient import BayEOSWriter, BayEOSSender
from ConfigParser import SafeConfigParser
from datetime import datetime, timedelta
import tempfile,urllib2,urllib,base64,requests 
import json
import os, sys, logging


log = logging.getLogger('root')

SEC_BRIDGE = 'tahmo'
SEC_GATEWAY = 'gateway'
ISO_TIMESTAMP = '%Y-%m-%dT%H:%M:%S'

minute = 60 
hour = 60 * minute
day = 24 * hour

userAgent = "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:54.0) Gecko/20100101 Firefox/54.0"


def main():
    conf = readConfig()        
    log.setLevel(str(conf.get(SEC_BRIDGE,'log_level')).upper())       
   
    pa = conf.get(SEC_BRIDGE,'path')           
    writer = BayEOSWriter(pa,max_chunk = conf.getint(SEC_BRIDGE, 'max_chunk_size'))              
    sender = BayEOSSender(pa,SEC_BRIDGE, conf.get(SEC_GATEWAY,'url') + conf.get(SEC_GATEWAY,'context_save'), conf.get(SEC_GATEWAY,'password'), conf.get(SEC_GATEWAY,'username'))
    sender.start()

    # Source 
    host = conf.get(SEC_BRIDGE,'host')
    sleep_time = conf.getint(SEC_BRIDGE,'sleep_time')
    
    # Destination
    gw_url = conf.get(SEC_GATEWAY,'url')
    gw_username = conf.get(SEC_GATEWAY,'username')
    gw_password = conf.get(SEC_GATEWAY,'password')

    devices = getDeviceList(conf)                              
    tahmo = [conf.get(SEC_BRIDGE,'host'),conf.get(SEC_BRIDGE,'username'),conf.get(SEC_BRIDGE,'password')]
    lrt_map = getLrtMap(gw_url,gw_username,gw_password,devices)     

    
    while True:  
        for device in devices:                                  
                    host, section, name = device                                                                                                 
                    origin = host + '/' + name
                    try:                                                      
                        if origin in lrt_map:                                                       
                            lrt = lrt_map[origin]
                            if lrt is None:
                                log.info("Initial import of {0}".format(origin))
                                if conf.has_option(SEC_BRIDGE,'initial_since'):
                                    since = datetime.strptime(conf.get(SEC_BRIDGE,'initial_since'),ISO_TIMESTAMP)                                                                                                                                                   
                                else:
                                    since = datetime.utcnow()
                            else:
                                log.info("Delta import of {0}.".format(origin))                             
                                since = lrt                             
                            f_date = fetchAndSaveData(tahmo,writer,device,since)
                            if f_date is not None:
                                lrt_map[origin] = f_date
                        else:                            
                            log.warn("Try to get last result time of {0} from gateway.".format(origin))
                            lrt_map[origin] = getLastResultTime(gw_url + conf.get(SEC_GATEWAY,'context_board'), gw_username, gw_password, origin) 

                    except urllib2.HTTPError as err:
                        log.error("HTTP communication error:{0}".format(str(err)))                        
                    except urllib2.URLError as err:
                        log.error('URLError: ' + str(err))
                    except ValueError as err:
                        log.error("Value error:{0}".format(str(err)))
                    except:
                        log.error('Unspecified error.' )
                    else: 
                        log.info("Import of {0} finished.".format(origin))
        
        log.info("Going to sleep for {0} secs ...".format(sleep_time))
        sleep(sleep_time)

def getLrtMap(url,username,password,devices):   
    """
    @param url: gateway url 
    @param username: gateway user name 
    @param password: gateway user password
    @param device: list of device tuples      
    """      
    log.info("Polling gateway for last result time values ...")
    lrt_map = {}    
    for device in devices:
        host, section, name = device
        origin = host + '/' + name
        try:
            lrt_map[origin] = getLastResultTime(url,username,password,origin)
        except:
            log.warning('Failed to get last result time of {0} from gateway.'.format(origin) )
    return lrt_map                        
    
def getDeviceList(conf):
    devices = []    
    host = conf.get(SEC_BRIDGE,'host')
    for section in conf.sections():
        if section.startswith('device'):
            devices.append((host, conf.get(section,'section'),conf.get(section,'name')))
    return devices           
    
def fetchAndSaveData(tahmo,writer,device,since):
    """ Fetch and save records from tahmo host      
    @param tahmo:  tahmo tuple
    @param writer: BayEOS Writer
    @param device: device tuple
    @param since: datetime
    @returns lrt: as datetime
    """    
    host, section, name = device
    thost, username, password = tahmo

    source_login_url =  "http://{0}/accounts/login/?next=/".format(host)
    source_data_url = "http://{0}/section/get_sectional_chart_info?".format(host)



    # Session
    session = requests.Session()
    headers = {'user-agent':userAgent}
    # Get cookie
    r = session.get(source_login_url,headers=headers)
    log.info("Cookie Request:{0}".format(r.status_code))
    token = r.cookies['csrftoken']
    # Login with token
    data = {"username":username,"password":password,"csrfmiddlewaretoken":token}
    r = session.post(source_login_url,data=data,headers=headers)
    log.info("Login Request:{0}".format(r.status_code))

    # Get data 
    param = {'section':section,'one_device':name, 'selected_id':'selected_id:-1', 'use_custom_period':'1'}
    
    time = since
    end_time = datetime.utcnow()
    lrt = None
    rc = 0
    while time <= end_time:        
        time = time + timedelta(days=1)
        param['custom_start'] = getDateString(time - timedelta(days=1))
        param['custom_end'] = getDateString(time)
        log.info("Query Device:{0} Start:{1} End:{2}".format(name,param['custom_start'],param['custom_end']))   

        r = session.get(source_data_url + urllib.urlencode(param),headers=headers)
        j = json.loads(r.text)
        data = j[3]['dataProvider']           
        for obs in data:  
           values = {}  
           for key, item in obs.items():
               if key == 'date':
                   ts = float(item)/1000   
                   lrt = datetime.utcfromtimestamp(ts)                 
               else:
                   values[key] = float(item)                                
           writer.save(values,value_type=0x61,timestamp=ts,origin=host + '/' + name) 
           rc = rc + 1;
        log.info("Sleeping 10 minutes...")   
        sleep(10*minute)
    log.info("{0} rows saved.".format(str(rc)))
    return lrt
    
def getDateString(dateTime):
    return str(dateTime.month - 1) + "/" + str(dateTime.day) + ",/" + str(dateTime.year) 

def getLastResultTime(url,username,password,origin):
    """
    Get the last result time value for channel on a BayEOS Gateway 
    @param url: gateway url
    @param username: gateway user name 
    @param password: gateway password
    @param origin: board origin   
    @return last result time as datetime             
    """               
    r = requests.get(url,params={'origin':origin}, auth=(username,password))
    if r.status_code == 404:
        log.warning("Origin:{0} not found.".format(origin))            
        return None        
    t = r.json()['lastResultTime'] 
    return datetime.strptime(t[:-1],'%Y-%m-%dT%H:%M:%S') 

def readConfig():
    """
        Reads the config file and populates default values 
        @return SafeConfigParser
    """
    conf = SafeConfigParser()
    if sys.platform == 'win32':        
        conf.read('tahmo.conf')
    else:
        conf.read('/etc/tahmo.conf')
    
    if not conf.has_option(SEC_BRIDGE, 'path'):
        conf.set(SEC_BRIDGE, 'path', os.path.join(tempfile.gettempdir(),SEC_BRIDGE))
    if not conf.has_option(SEC_BRIDGE, 'max_chunk_size'):
        conf.set(SEC_BRIDGE, 'max_chunk_size', "32000")
    if not conf.has_option(SEC_GATEWAY, 'context_board'):
        conf.set(SEC_GATEWAY, 'context_board', "/board/findByOrigin")
    if not conf.has_option(SEC_GATEWAY, 'context_save'):
        conf.set(SEC_GATEWAY, 'context_save', "/frame/saveFlat")
        

    return conf
        
if __name__ == '__main__':
    main()